# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from datetime import datetime
from typing import Union, TYPE_CHECKING

from bungio.utils import enum_converter
from bungio.models.base import BaseModel, custom_define, custom_field

from bungio.models.mixins import DestinyUserMixin

if TYPE_CHECKING:
    from bungio.models import SingleComponentResponseOfDestinySocialCommendationsComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyPresentationNodesComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyInventoryComponent
    from bungio.models import SingleComponentResponseOfDestinyCharacterActivitiesComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyCurrenciesComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyCharacterProgressionComponent
    from bungio.models import SingleComponentResponseOfDestinyCollectiblesComponent
    from bungio.models import SingleComponentResponseOfDestinyCharacterProgressionComponent
    from bungio.models import SingleComponentResponseOfDestinyProfileProgressionComponent
    from bungio.models import SingleComponentResponseOfDestinyKiosksComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyCharacterRecordsComponent
    from bungio.models import SingleComponentResponseOfDestinyItemObjectivesComponent
    from bungio.models import SingleComponentResponseOfDestinyItemPerksComponent
    from bungio.models import SingleComponentResponseOfDestinyVendorComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyLoadoutsComponent
    from bungio.models import DictionaryComponentResponseOfint32AndDestinyVendorSaleItemComponent
    from bungio.models import DestinyItemComponentSetOfint64
    from bungio.models import DestinyVendorSaleItemComponent
    from bungio.models import SingleComponentResponseOfDestinyPresentationNodesComponent
    from bungio.models import SingleComponentResponseOfDestinyPlugSetsComponent
    from bungio.models import SingleComponentResponseOfDestinyItemComponent
    from bungio.models import PlatformErrorCodes
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyKiosksComponent
    from bungio.models import DestinyItemComponent
    from bungio.models import DestinyItemComponentSetOfuint32
    from bungio.models import SingleComponentResponseOfDestinyProfileRecordsComponent
    from bungio.models import SingleComponentResponseOfDestinyVendorReceiptsComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyCharacterComponent
    from bungio.models import SingleComponentResponseOfDestinyStringVariablesComponent
    from bungio.models import DestinyPlatformSilverComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyCollectiblesComponent
    from bungio.models import SingleComponentResponseOfDestinyCharacterRecordsComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyPlugSetsComponent
    from bungio.models import UserInfoCard
    from bungio.models import SingleComponentResponseOfDestinyProfileComponent
    from bungio.models import SingleComponentResponseOfDestinyLoadoutsComponent
    from bungio.models import SingleComponentResponseOfDestinyItemSocketsComponent
    from bungio.models import BungieMembershipType
    from bungio.models import SingleComponentResponseOfDestinyItemPlugObjectivesComponent
    from bungio.models import SingleComponentResponseOfDestinyCurrenciesComponent
    from bungio.models import DestinyItemComponentSetOfint32
    from bungio.models import SingleComponentResponseOfDestinyPlatformSilverComponent
    from bungio.models import DictionaryComponentResponseOfuint32AndDestinyVendorCategoriesComponent
    from bungio.models import SingleComponentResponseOfDestinyProfileCollectiblesComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyCraftablesComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyStringVariablesComponent
    from bungio.models import DictionaryComponentResponseOfuint32AndPersonalDestinyVendorSaleItemSetComponent
    from bungio.models import DictionaryComponentResponseOfuint32AndDestinyPublicVendorComponent
    from bungio.models import SingleComponentResponseOfDestinyMetricsComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyCharacterActivitiesComponent
    from bungio.models import DictionaryComponentResponseOfuint32AndPublicDestinyVendorSaleItemSetComponent
    from bungio.models import DestinyBaseItemComponentSetOfuint32
    from bungio.models import SingleComponentResponseOfDestinyItemReusablePlugsComponent
    from bungio.models import SingleComponentResponseOfDestinyVendorCategoriesComponent
    from bungio.models import SingleComponentResponseOfDestinyItemTalentGridComponent
    from bungio.models import DictionaryComponentResponseOfuint32AndDestinyVendorComponent
    from bungio.models import SingleComponentResponseOfDestinyCharacterComponent
    from bungio.models import SingleComponentResponseOfDestinyVendorGroupComponent
    from bungio.models import DictionaryComponentResponseOfint64AndDestinyCharacterRenderComponent
    from bungio.models import SingleComponentResponseOfDestinyCharacterRenderComponent
    from bungio.models import SingleComponentResponseOfDestinyInventoryComponent
    from bungio.models import SingleComponentResponseOfDestinyItemStatsComponent
    from bungio.models import SingleComponentResponseOfDestinyProfileTransitoryComponent
    from bungio.models import DestinyPublicVendorSaleItemComponent
    from bungio.models import SingleComponentResponseOfDestinyItemInstanceComponent
    from bungio.models import SingleComponentResponseOfDestinyItemRenderComponent


@custom_define()
class DestinyLinkedProfilesResponse(BaseModel):
    """
    I know what you seek. You seek linked accounts. Found them, you have. This contract returns a minimal amount of data about Destiny Accounts that are linked through your Bungie.Net account. We will not return accounts in this response whose

    None
    Attributes:
        bnet_membership: If the requested membership had a linked Bungie.Net membership ID, this is the basic information about that BNet account. I know, Tetron; I know this is mixing UserServices concerns with DestinyServices concerns. But it's so damn convenient! https://www.youtube.com/watch?v=X5R-bB-gKVI
        profiles: Any Destiny account for whom we could successfully pull characters will be returned here, as the Platform-level summary of user data. (no character data, no Destiny account data other than the Membership ID and Type so you can make further queries)
        profiles_with_errors: This is brief summary info for profiles that we believe have valid Destiny info, but who failed to return data for some other reason and thus we know that subsequent calls for their info will also fail.
    """

    bnet_membership: "UserInfoCard" = custom_field()
    profiles: list["DestinyProfileUserInfoCard"] = custom_field(
        metadata={"type": """list[DestinyProfileUserInfoCard]"""}
    )
    profiles_with_errors: list["DestinyErrorProfile"] = custom_field(metadata={"type": """list[DestinyErrorProfile]"""})


@custom_define()
class DestinyProfileUserInfoCard(BaseModel, DestinyUserMixin):
    """
    _No description given by bungie._

    None
    Attributes:
        applicable_membership_types: The list of Membership Types indicating the platforms on which this Membership can be used.  Not in Cross Save = its original membership type. Cross Save Primary = Any membership types it is overridding, and its original membership type Cross Save Overridden = Empty list
        bungie_global_display_name: The bungie global display name, if set.
        bungie_global_display_name_code: The bungie global display name code, if set.
        cross_save_override: If there is a cross save override in effect, this value will tell you the type that is overridding this one.
        date_last_played: _No description given by bungie._
        display_name: Display Name the player has chosen for themselves. The display name is optional when the data type is used as input to a platform API.
        icon_path: URL the Icon if available.
        is_cross_save_primary: If true, this account is hooked up as the "Primary" cross save account for one or more platforms.
        is_overridden: If this profile is being overridden/obscured by Cross Save, this will be set to true. We will still return the profile for display purposes where users need to know the info: it is up to any given area of the app/site to determine if this profile should still be shown.
        is_public: If True, this is a public user membership.
        membership_id: Membership ID as they user is known in the Accounts service
        membership_type: Type of the membership. Not necessarily the native type.
        platform_silver: This is the silver available on this Profile across any platforms on which they have purchased silver.  This is only available if you are requesting yourself.
        supplemental_display_name: A platform specific additional display name - ex: psn Real Name, bnet Unique Name, etc.
        unpaired_game_versions: If this profile is not in a cross save pairing, this will return the game versions that we believe this profile has access to.  For the time being, we will not return this information for any membership that is in a cross save pairing. The gist is that, once the pairing occurs, we do not currently have a consistent way to get that information for the profile's original Platform, and thus gameVersions would be too inconsistent (based on the last platform they happened to play on) for the info to be useful.  If we ever can get this data, this field will be deprecated and replaced with data on the DestinyLinkedProfileResponse itself, with game versions per linked Platform. But since we can't get that, we have this as a stop-gap measure for getting the data in the only situation that we currently need it.
    """

    applicable_membership_types: list[Union["BungieMembershipType", int]] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    bungie_global_display_name: str = custom_field()
    bungie_global_display_name_code: int = custom_field()
    cross_save_override: Union["BungieMembershipType", int] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    date_last_played: datetime = custom_field()
    display_name: str = custom_field()
    icon_path: str = custom_field()
    is_cross_save_primary: bool = custom_field()
    is_overridden: bool = custom_field()
    is_public: bool = custom_field()
    membership_id: int = custom_field(metadata={"int64": True})
    membership_type: Union["BungieMembershipType", int] = custom_field(converter=enum_converter("BungieMembershipType"))
    platform_silver: "DestinyPlatformSilverComponent" = custom_field()
    supplemental_display_name: str = custom_field()
    unpaired_game_versions: int = custom_field()


@custom_define()
class DestinyErrorProfile(BaseModel):
    """
    If a Destiny Profile can't be returned, but we're pretty certain it's a valid Destiny account, this will contain as much info as we can get about the profile for your use. Assume that the most you'll get is the Error Code, the Membership Type and the Membership ID.

    None
    Attributes:
        error_code: The error that we encountered. You should be able to look up localized text to show to the user for these failures.
        info_card: Basic info about the account that failed. Don't expect anything other than membership ID, Membership Type, and displayName to be populated.
    """

    error_code: Union["PlatformErrorCodes", int] = custom_field(converter=enum_converter("PlatformErrorCodes"))
    info_card: "UserInfoCard" = custom_field()


@custom_define()
class DestinyProfileResponse(BaseModel):
    """
    The response for GetDestinyProfile, with components for character and item-level data.

    None
    Attributes:
        character_activities: Character activity data - the activities available to this character and its status, keyed by the Character's Id. COMPONENT TYPE: CharacterActivities
        character_collectibles: COMPONENT TYPE: Collectibles
        character_craftables: COMPONENT TYPE: Craftables
        character_currency_lookups: A "lookup" convenience component that can be used to quickly check if the character has access to items that can be used for purchasing. COMPONENT TYPE: CurrencyLookups
        character_equipment: The character's equipped items, keyed by the Character's Id. COMPONENT TYPE: CharacterEquipment
        character_inventories: The character-level non-equipped inventory items, keyed by the Character's Id. COMPONENT TYPE: CharacterInventories
        character_kiosks: Items available from Kiosks that are available to a specific character as opposed to the account as a whole. It must be combined with data from the profileKiosks property to get a full picture of the character's available items to check out of a kiosk. This component returns information about what Kiosk items are available to you on a *Character* level. Usually, kiosk items will be earned for the entire Profile (all characters) at once. To find those, look in the profileKiosks property. COMPONENT TYPE: Kiosks
        character_loadouts: The character loadouts, keyed by the Character's Id. COMPONENT TYPE: CharacterLoadouts
        character_plug_sets: When sockets refer to reusable Plug Sets (see DestinyPlugSetDefinition for more info), this is the set of plugs and their states, per character, that are character-scoped. This comes back with ItemSockets, as it is needed for a complete picture of the sockets on requested items. COMPONENT TYPE: ItemSockets
        character_presentation_nodes: COMPONENT TYPE: PresentationNodes
        character_progressions: Character-level progression data, keyed by the Character's Id. COMPONENT TYPE: CharacterProgressions
        character_records: COMPONENT TYPE: Records
        character_render_data: Character rendering data - a minimal set of info needed to render a character in 3D - keyed by the Character's Id. COMPONENT TYPE: CharacterRenderData
        character_string_variables: COMPONENT TYPE: StringVariables
        character_uninstanced_item_components: Do you ever get the feeling that a system was designed *too* flexibly? That it can be used in so many different ways that you end up being unable to provide an easy to use abstraction for the mess that's happening under the surface? Let's talk about character-specific data that might be related to items without instances. These two statements are totally unrelated, I promise. At some point during D2, it was decided that items - such as Bounties - could be given to characters and *not* have instance data, but that *could* display and even use relevant state information on your account and character. Up to now, any item that had meaningful dependencies on character or account state had to be instanced, and thus "itemComponents" was all that you needed: it was keyed by item's instance IDs and provided the stateful information you needed inside. Unfortunately, we don't live in such a magical world anymore. This is information held on a per-character basis about non-instanced items that the characters have in their inventory - or that reference character-specific state information even if it's in Account-level inventory - and the values related to that item's state in relation to the given character. To give a concrete example, look at a Moments of Triumph bounty. They exist in a character's inventory, and show/care about a character's progression toward completing the bounty. But the bounty itself is a non-instanced item, like a mod or a currency. This returns that data for the characters who have the bounty in their inventory. I'm not crying, you're crying Okay we're both crying but it's going to be okay I promise Actually I shouldn't promise that, I don't know if it's going to be okay
        characters: Basic information about each character, keyed by the CharacterId. COMPONENT TYPE: Characters
        item_components: Information about instanced items across all returned characters, keyed by the item's instance ID. COMPONENT TYPE: [See inside the DestinyItemComponentSet contract for component types.]
        metrics: COMPONENT TYPE: Metrics
        platform_silver: Silver quantities for any platform on which this Profile plays destiny.  COMPONENT TYPE: PlatformSilver
        profile: The basic information about the Destiny Profile (formerly "Account"). COMPONENT TYPE: Profiles
        profile_collectibles: COMPONENT TYPE: Collectibles
        profile_commendations: COMPONENT TYPE: SocialCommendations
        profile_currencies: The profile-level currencies owned by the Destiny Profile. COMPONENT TYPE: ProfileCurrencies
        profile_inventory: The profile-level inventory of the Destiny Profile. COMPONENT TYPE: ProfileInventories
        profile_kiosks: Items available from Kiosks that are available Profile-wide (i.e. across all characters) This component returns information about what Kiosk items are available to you on a *Profile* level. It is theoretically possible for Kiosks to have items gated by specific Character as well. If you ever have those, you will find them on the characterKiosks property. COMPONENT TYPE: Kiosks
        profile_plug_sets: When sockets refer to reusable Plug Sets (see DestinyPlugSetDefinition for more info), this is the set of plugs and their states that are profile-scoped. This comes back with ItemSockets, as it is needed for a complete picture of the sockets on requested items. COMPONENT TYPE: ItemSockets
        profile_presentation_nodes: COMPONENT TYPE: PresentationNodes
        profile_progression: When we have progression information - such as Checklists - that may apply profile-wide, it will be returned here rather than in the per-character progression data. COMPONENT TYPE: ProfileProgression
        profile_records: COMPONENT TYPE: Records
        profile_string_variables: COMPONENT TYPE: StringVariables
        profile_transitory_data: COMPONENT TYPE: Transitory
        response_minted_timestamp: Records the timestamp of when most components were last generated from the world server source. Unless the component type is specified in the documentation for secondaryComponentsMintedTimestamp, this value is sufficient to do data freshness.
        secondary_components_minted_timestamp: Some secondary components are not tracked in the primary response timestamp and have their timestamp tracked here. If your component is any of the following, this field is where you will find your timestamp value:  PresentationNodes, Records, Collectibles, Metrics, StringVariables, Craftables, Transitory  All other component types may use the primary timestamp property.
        vendor_receipts: Recent, refundable purchases you have made from vendors. When will you use it? Couldn't say... COMPONENT TYPE: VendorReceipts
    """

    character_activities: "DictionaryComponentResponseOfint64AndDestinyCharacterActivitiesComponent" = custom_field()
    character_collectibles: "DictionaryComponentResponseOfint64AndDestinyCollectiblesComponent" = custom_field()
    character_craftables: "DictionaryComponentResponseOfint64AndDestinyCraftablesComponent" = custom_field()
    character_currency_lookups: "DictionaryComponentResponseOfint64AndDestinyCurrenciesComponent" = custom_field()
    character_equipment: "DictionaryComponentResponseOfint64AndDestinyInventoryComponent" = custom_field()
    character_inventories: "DictionaryComponentResponseOfint64AndDestinyInventoryComponent" = custom_field()
    character_kiosks: "DictionaryComponentResponseOfint64AndDestinyKiosksComponent" = custom_field()
    character_loadouts: "DictionaryComponentResponseOfint64AndDestinyLoadoutsComponent" = custom_field()
    character_plug_sets: "DictionaryComponentResponseOfint64AndDestinyPlugSetsComponent" = custom_field()
    character_presentation_nodes: "DictionaryComponentResponseOfint64AndDestinyPresentationNodesComponent" = (
        custom_field()
    )
    character_progressions: "DictionaryComponentResponseOfint64AndDestinyCharacterProgressionComponent" = custom_field()
    character_records: "DictionaryComponentResponseOfint64AndDestinyCharacterRecordsComponent" = custom_field()
    character_render_data: "DictionaryComponentResponseOfint64AndDestinyCharacterRenderComponent" = custom_field()
    character_string_variables: "DictionaryComponentResponseOfint64AndDestinyStringVariablesComponent" = custom_field()
    character_uninstanced_item_components: dict[int, "DestinyBaseItemComponentSetOfuint32"] = custom_field(
        metadata={"type": """dict[int, DestinyBaseItemComponentSetOfuint32]"""}
    )
    characters: "DictionaryComponentResponseOfint64AndDestinyCharacterComponent" = custom_field()
    item_components: "DestinyItemComponentSetOfint64" = custom_field()
    metrics: "SingleComponentResponseOfDestinyMetricsComponent" = custom_field()
    platform_silver: "SingleComponentResponseOfDestinyPlatformSilverComponent" = custom_field()
    profile: "SingleComponentResponseOfDestinyProfileComponent" = custom_field()
    profile_collectibles: "SingleComponentResponseOfDestinyProfileCollectiblesComponent" = custom_field()
    profile_commendations: "SingleComponentResponseOfDestinySocialCommendationsComponent" = custom_field()
    profile_currencies: "SingleComponentResponseOfDestinyInventoryComponent" = custom_field()
    profile_inventory: "SingleComponentResponseOfDestinyInventoryComponent" = custom_field()
    profile_kiosks: "SingleComponentResponseOfDestinyKiosksComponent" = custom_field()
    profile_plug_sets: "SingleComponentResponseOfDestinyPlugSetsComponent" = custom_field()
    profile_presentation_nodes: "SingleComponentResponseOfDestinyPresentationNodesComponent" = custom_field()
    profile_progression: "SingleComponentResponseOfDestinyProfileProgressionComponent" = custom_field()
    profile_records: "SingleComponentResponseOfDestinyProfileRecordsComponent" = custom_field()
    profile_string_variables: "SingleComponentResponseOfDestinyStringVariablesComponent" = custom_field()
    profile_transitory_data: "SingleComponentResponseOfDestinyProfileTransitoryComponent" = custom_field()
    response_minted_timestamp: datetime = custom_field()
    secondary_components_minted_timestamp: datetime = custom_field()
    vendor_receipts: "SingleComponentResponseOfDestinyVendorReceiptsComponent" = custom_field()


@custom_define()
class DestinyCharacterResponse(BaseModel):
    """
    The response contract for GetDestinyCharacter, with components that can be returned for character and item-level data.

    None
    Attributes:
        activities: Activity data - info about current activities available to the player. COMPONENT TYPE: CharacterActivities
        character: Base information about the character in question. COMPONENT TYPE: Characters
        collectibles: COMPONENT TYPE: Collectibles
        currency_lookups: A "lookup" convenience component that can be used to quickly check if the character has access to items that can be used for purchasing. COMPONENT TYPE: CurrencyLookups
        equipment: Equipped items on the character. COMPONENT TYPE: CharacterEquipment
        inventory: The character-level non-equipped inventory items. COMPONENT TYPE: CharacterInventories
        item_components: The set of components belonging to the player's instanced items. COMPONENT TYPE: [See inside the DestinyItemComponentSet contract for component types.]
        kiosks: Items available from Kiosks that are available to this specific character.  COMPONENT TYPE: Kiosks
        loadouts: The loadouts available to the character. COMPONENT TYPE: CharacterLoadouts
        plug_sets: When sockets refer to reusable Plug Sets (see DestinyPlugSetDefinition for more info), this is the set of plugs and their states that are scoped to this character. This comes back with ItemSockets, as it is needed for a complete picture of the sockets on requested items. COMPONENT TYPE: ItemSockets
        presentation_nodes: COMPONENT TYPE: PresentationNodes
        progressions: Character progression data, including Milestones. COMPONENT TYPE: CharacterProgressions
        records: COMPONENT TYPE: Records
        render_data: Character rendering data - a minimal set of information about equipment and dyes used for rendering. COMPONENT TYPE: CharacterRenderData
        uninstanced_item_components: The set of components belonging to the player's UNinstanced items. Because apparently now those too can have information relevant to the character's state. COMPONENT TYPE: [See inside the DestinyItemComponentSet contract for component types.]
    """

    activities: "SingleComponentResponseOfDestinyCharacterActivitiesComponent" = custom_field()
    character: "SingleComponentResponseOfDestinyCharacterComponent" = custom_field()
    collectibles: "SingleComponentResponseOfDestinyCollectiblesComponent" = custom_field()
    currency_lookups: "SingleComponentResponseOfDestinyCurrenciesComponent" = custom_field()
    equipment: "SingleComponentResponseOfDestinyInventoryComponent" = custom_field()
    inventory: "SingleComponentResponseOfDestinyInventoryComponent" = custom_field()
    item_components: "DestinyItemComponentSetOfint64" = custom_field()
    kiosks: "SingleComponentResponseOfDestinyKiosksComponent" = custom_field()
    loadouts: "SingleComponentResponseOfDestinyLoadoutsComponent" = custom_field()
    plug_sets: "SingleComponentResponseOfDestinyPlugSetsComponent" = custom_field()
    presentation_nodes: "SingleComponentResponseOfDestinyPresentationNodesComponent" = custom_field()
    progressions: "SingleComponentResponseOfDestinyCharacterProgressionComponent" = custom_field()
    records: "SingleComponentResponseOfDestinyCharacterRecordsComponent" = custom_field()
    render_data: "SingleComponentResponseOfDestinyCharacterRenderComponent" = custom_field()
    uninstanced_item_components: "DestinyBaseItemComponentSetOfuint32" = custom_field()


@custom_define()
class DestinyItemResponse(BaseModel):
    """
    The response object for retrieving an individual instanced item. None of these components are relevant for an item that doesn't have an "itemInstanceId": for those, get your information from the DestinyInventoryDefinition.

    None
    Attributes:
        character_id: If the item is on a character, this will return the ID of the character that is holding the item.
        instance: Basic instance data for the item. COMPONENT TYPE: ItemInstances
        item: Common data for the item relevant to its non-instanced properties. COMPONENT TYPE: ItemCommonData
        objectives: Information specifically about the item's objectives. COMPONENT TYPE: ItemObjectives
        perks: Information specifically about the perks currently active on the item. COMPONENT TYPE: ItemPerks
        plug_objectives: Information about objectives on Plugs for a given item. See the component's documentation for more info. COMPONENT TYPE: ItemPlugObjectives
        render_data: Information about how to render the item in 3D. COMPONENT TYPE: ItemRenderData
        reusable_plugs: Information about the Reusable Plugs for sockets on an item. These are plugs that you can insert into the given socket regardless of if you actually own an instance of that plug: they are logic-driven plugs rather than inventory-driven.  These may need to be combined with Plug Set component data to get a full picture of available plugs on a given socket.  COMPONENT TYPE: ItemReusablePlugs
        sockets: Information about the sockets of the item: which are currently active, what potential sockets you could have and the stats/abilities/perks you can gain from them. COMPONENT TYPE: ItemSockets
        stats: Information about the computed stats of the item: power, defense, etc... COMPONENT TYPE: ItemStats
        talent_grid: Information about the talent grid attached to the item. Talent nodes can provide a variety of benefits and abilities, and in Destiny 2 are used almost exclusively for the character's "Builds". COMPONENT TYPE: ItemTalentGrids
    """

    character_id: int = custom_field(metadata={"int64": True})
    instance: "SingleComponentResponseOfDestinyItemInstanceComponent" = custom_field()
    item: "SingleComponentResponseOfDestinyItemComponent" = custom_field()
    objectives: "SingleComponentResponseOfDestinyItemObjectivesComponent" = custom_field()
    perks: "SingleComponentResponseOfDestinyItemPerksComponent" = custom_field()
    plug_objectives: "SingleComponentResponseOfDestinyItemPlugObjectivesComponent" = custom_field()
    render_data: "SingleComponentResponseOfDestinyItemRenderComponent" = custom_field()
    reusable_plugs: "SingleComponentResponseOfDestinyItemReusablePlugsComponent" = custom_field()
    sockets: "SingleComponentResponseOfDestinyItemSocketsComponent" = custom_field()
    stats: "SingleComponentResponseOfDestinyItemStatsComponent" = custom_field()
    talent_grid: "SingleComponentResponseOfDestinyItemTalentGridComponent" = custom_field()


@custom_define()
class DestinyVendorsResponse(BaseModel):
    """
    A response containing all of the components for all requested vendors.

    None
    Attributes:
        categories: Categories that the vendor has available, and references to the sales therein. These are keyed by the Vendor Hash, so you will get one Categories Component per vendor returned. COMPONENT TYPE: VendorCategories
        currency_lookups: A "lookup" convenience component that can be used to quickly check if the character has access to items that can be used for purchasing. COMPONENT TYPE: CurrencyLookups
        item_components: The set of item detail components, one set of item components per Vendor. These are keyed by the Vendor Hash, so you will get one Item Component Set per vendor returned. The components contained inside are themselves keyed by the vendorSaleIndex, and will have whatever item-level components you requested (Sockets, Stats, Instance data etc...) per item being sold by the vendor.
        sales: Sales, keyed by the vendorItemIndex of the item being sold. These are keyed by the Vendor Hash, so you will get one Sale Item Set Component per vendor returned. Note that within the Sale Item Set component, the sales are themselves keyed by the vendorSaleIndex, so you can relate it to the corrent sale item definition within the Vendor's definition. COMPONENT TYPE: VendorSales
        string_variables: A map of string variable values by hash for this character context. COMPONENT TYPE: StringVariables
        vendor_groups: For Vendors being returned, this will give you the information you need to group them and order them in the same way that the Bungie Companion app performs grouping. It will automatically be returned if you request the Vendors component. COMPONENT TYPE: Vendors
        vendors: The base properties of the vendor. These are keyed by the Vendor Hash, so you will get one Vendor Component per vendor returned. COMPONENT TYPE: Vendors
    """

    categories: "DictionaryComponentResponseOfuint32AndDestinyVendorCategoriesComponent" = custom_field()
    currency_lookups: "SingleComponentResponseOfDestinyCurrenciesComponent" = custom_field()
    item_components: dict[int, "DestinyItemComponentSetOfint32"] = custom_field(
        metadata={"type": """dict[int, DestinyItemComponentSetOfint32]"""}
    )
    sales: "DictionaryComponentResponseOfuint32AndPersonalDestinyVendorSaleItemSetComponent" = custom_field()
    string_variables: "SingleComponentResponseOfDestinyStringVariablesComponent" = custom_field()
    vendor_groups: "SingleComponentResponseOfDestinyVendorGroupComponent" = custom_field()
    vendors: "DictionaryComponentResponseOfuint32AndDestinyVendorComponent" = custom_field()


@custom_define()
class PersonalDestinyVendorSaleItemSetComponent(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        sale_items: _No description given by bungie._
    """

    sale_items: dict[int, "DestinyVendorSaleItemComponent"] = custom_field(
        metadata={"type": """dict[int, DestinyVendorSaleItemComponent]"""}
    )


@custom_define()
class DestinyVendorResponse(BaseModel):
    """
    A response containing all of the components for a vendor.

    None
    Attributes:
        categories: Categories that the vendor has available, and references to the sales therein. COMPONENT TYPE: VendorCategories
        currency_lookups: A "lookup" convenience component that can be used to quickly check if the character has access to items that can be used for purchasing. COMPONENT TYPE: CurrencyLookups
        item_components: Item components, keyed by the vendorItemIndex of the active sale items. COMPONENT TYPE: [See inside the DestinyItemComponentSet contract for component types.]
        sales: Sales, keyed by the vendorItemIndex of the item being sold. COMPONENT TYPE: VendorSales
        string_variables: A map of string variable values by hash for this character context. COMPONENT TYPE: StringVariables
        vendor: The base properties of the vendor. COMPONENT TYPE: Vendors
    """

    categories: "SingleComponentResponseOfDestinyVendorCategoriesComponent" = custom_field()
    currency_lookups: "SingleComponentResponseOfDestinyCurrenciesComponent" = custom_field()
    item_components: "DestinyItemComponentSetOfint32" = custom_field()
    sales: "DictionaryComponentResponseOfint32AndDestinyVendorSaleItemComponent" = custom_field()
    string_variables: "SingleComponentResponseOfDestinyStringVariablesComponent" = custom_field()
    vendor: "SingleComponentResponseOfDestinyVendorComponent" = custom_field()


@custom_define()
class DestinyPublicVendorsResponse(BaseModel):
    """
    A response containing all valid components for the public Vendors endpoint.  It is a decisively smaller subset of data compared to what we can get when we know the specific user making the request.  If you want any of the other data - item details, whether or not you can buy it, etc... you'll have to call in the context of a character. I know, sad but true.

    None
    Attributes:
        categories: Categories that the vendor has available, and references to the sales therein. These are keyed by the Vendor Hash, so you will get one Categories Component per vendor returned. COMPONENT TYPE: VendorCategories
        sales: Sales, keyed by the vendorItemIndex of the item being sold. These are keyed by the Vendor Hash, so you will get one Sale Item Set Component per vendor returned. Note that within the Sale Item Set component, the sales are themselves keyed by the vendorSaleIndex, so you can relate it to the corrent sale item definition within the Vendor's definition. COMPONENT TYPE: VendorSales
        string_variables: A set of string variable values by hash for a public vendors context. COMPONENT TYPE: StringVariables
        vendor_groups: For Vendors being returned, this will give you the information you need to group them and order them in the same way that the Bungie Companion app performs grouping. It will automatically be returned if you request the Vendors component. COMPONENT TYPE: Vendors
        vendors: The base properties of the vendor. These are keyed by the Vendor Hash, so you will get one Vendor Component per vendor returned. COMPONENT TYPE: Vendors
    """

    categories: "DictionaryComponentResponseOfuint32AndDestinyVendorCategoriesComponent" = custom_field()
    sales: "DictionaryComponentResponseOfuint32AndPublicDestinyVendorSaleItemSetComponent" = custom_field()
    string_variables: "SingleComponentResponseOfDestinyStringVariablesComponent" = custom_field()
    vendor_groups: "SingleComponentResponseOfDestinyVendorGroupComponent" = custom_field()
    vendors: "DictionaryComponentResponseOfuint32AndDestinyPublicVendorComponent" = custom_field()


@custom_define()
class PublicDestinyVendorSaleItemSetComponent(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        sale_items: _No description given by bungie._
    """

    sale_items: dict[int, "DestinyPublicVendorSaleItemComponent"] = custom_field(
        metadata={"type": """dict[int, DestinyPublicVendorSaleItemComponent]"""}
    )


@custom_define()
class DestinyCollectibleNodeDetailResponse(BaseModel):
    """
    Returns the detailed information about a Collectible Presentation Node and any Collectibles that are direct descendants.

    None
    Attributes:
        collectible_item_components: Item components, keyed by the item hash of the items pointed at collectibles found under the requested Presentation Node. NOTE: I had a lot of hemming and hawing about whether these should be keyed by collectible hash or item hash... but ultimately having it be keyed by item hash meant that UI that already uses DestinyItemComponentSet data wouldn't have to have a special override to do the collectible -> item lookup once you delve into an item's details, and it also meant that you didn't have to remember that the Hash being used as the key for plugSets was different from the Hash being used for the other Dictionaries. As a result, using the Item Hash felt like the least crappy solution. We may all come to regret this decision. We will see. COMPONENT TYPE: [See inside the DestinyItemComponentSet contract for component types.]
        collectibles: COMPONENT TYPE: Collectibles
    """

    collectible_item_components: "DestinyItemComponentSetOfuint32" = custom_field()
    collectibles: "SingleComponentResponseOfDestinyCollectiblesComponent" = custom_field()


@custom_define()
class InventoryChangedResponse(BaseModel):
    """
    A response containing all of the components for all requested vendors.

    None
    Attributes:
        added_inventory_items: Items that appeared in the inventory possibly as a result of an action.
        removed_inventory_items: Items that disappeared from the inventory possibly as a result of an action.
    """

    added_inventory_items: list["DestinyItemComponent"] = custom_field(
        metadata={"type": """list[DestinyItemComponent]"""}
    )
    removed_inventory_items: list["DestinyItemComponent"] = custom_field(
        metadata={"type": """list[DestinyItemComponent]"""}
    )


@custom_define()
class DestinyItemChangeResponse(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        added_inventory_items: Items that appeared in the inventory possibly as a result of an action.
        item: _No description given by bungie._
        removed_inventory_items: Items that disappeared from the inventory possibly as a result of an action.
    """

    added_inventory_items: list["DestinyItemComponent"] = custom_field(
        metadata={"type": """list[DestinyItemComponent]"""}
    )
    item: "DestinyItemResponse" = custom_field()
    removed_inventory_items: list["DestinyItemComponent"] = custom_field(
        metadata={"type": """list[DestinyItemComponent]"""}
    )



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.


from bungio.models.base import BaseModel, custom_define, custom_field


@custom_define()
class SearchResult(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        has_more: _No description given by bungie._
        query: _No description given by bungie._
        replacement_continuation_token: _No description given by bungie._
        total_results: _No description given by bungie._
        use_total_results: If useTotalResults is true, then totalResults represents an accurate count. If False, it does not, and may be estimated/only the size of the current page. Either way, you should probably always only trust hasMore. This is a long-held historical throwback to when we used to do paging with known total results. Those queries toasted our database, and we were left to hastily alter our endpoints and create backward- compatible shims, of which useTotalResults is one.
    """

    has_more: bool = custom_field()
    query: "PagedQuery" = custom_field()
    replacement_continuation_token: str = custom_field()
    total_results: int = custom_field()
    use_total_results: bool = custom_field()


@custom_define()
class PagedQuery(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        current_page: _No description given by bungie._
        items_per_page: _No description given by bungie._
        request_continuation_token: _No description given by bungie._
    """

    current_page: int = custom_field()
    items_per_page: int = custom_field()
    request_continuation_token: str = custom_field()



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import TYPE_CHECKING

from bungio.models.base import BaseModel, custom_define, custom_field


if TYPE_CHECKING:
    from bungio.models import DestinyItemPlug


@custom_define()
class DestinyPlugSetsComponent(BaseModel):
    """
    Sockets may refer to a "Plug Set": a set of reusable plugs that may be shared across multiple sockets (or even, in theory, multiple sockets over multiple items). This is the set of those plugs that we came across in the users' inventory, along with the values for plugs in the set. Any given set in this component may be represented in Character and Profile-level, as some plugs may be Profile-level restricted, and some character-level restricted. (note that the ones that are even more specific will remain on the actual socket component itself, as they cannot be reused)

    None
    Attributes:
        plugs: The shared list of plugs for each relevant PlugSet, keyed by the hash identifier of the PlugSet (DestinyPlugSetDefinition).
    """

    plugs: dict[int, list["DestinyItemPlug"]] = custom_field(metadata={"type": """dict[int, list[DestinyItemPlug]]"""})



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from datetime import datetime
from typing import Optional, Any, Union, TYPE_CHECKING

from bungio.utils import enum_converter
from bungio.models.base import BaseModel, BaseEnum, BaseFlagEnum, HashObject, ManifestModel, custom_define, custom_field


if TYPE_CHECKING:
    from bungio.models import GeneralUser


@custom_define()
class ContentItemPublicContract(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        allow_comments: _No description given by bungie._
        author: _No description given by bungie._
        auto_english_property_fallback: _No description given by bungie._
        c_type: _No description given by bungie._
        cms_path: _No description given by bungie._
        comment_summary: _No description given by bungie._
        content_id: _No description given by bungie._
        creation_date: _No description given by bungie._
        has_age_gate: _No description given by bungie._
        minimum_age: _No description given by bungie._
        modify_date: _No description given by bungie._
        properties: Firehose content is really a collection of metadata and "properties", which are the potentially-but-not-strictly localizable data that comprises the meat of whatever content is being shown. As Cole Porter would have crooned, "Anything Goes" with Firehose properties. They are most often strings, but they can theoretically be anything. They are JSON encoded, and could be JSON structures, simple strings, numbers etc... The Content Type of the item (cType) will describe the properties, and thus how they ought to be deserialized.
        rating_image_path: _No description given by bungie._
        representations: _No description given by bungie._
        tags: NOTE: Tags will always be lower case.
    """

    allow_comments: bool = custom_field()
    author: "GeneralUser" = custom_field()
    auto_english_property_fallback: bool = custom_field()
    c_type: str = custom_field()
    cms_path: str = custom_field()
    comment_summary: "CommentSummary" = custom_field()
    content_id: int = custom_field(metadata={"int64": True})
    creation_date: datetime = custom_field()
    has_age_gate: bool = custom_field()
    minimum_age: int = custom_field()
    modify_date: datetime = custom_field()
    properties: dict[str, Any] = custom_field(metadata={"type": """dict[str, Any]"""})
    rating_image_path: str = custom_field()
    representations: list["ContentRepresentation"] = custom_field(metadata={"type": """list[ContentRepresentation]"""})
    tags: list[str] = custom_field(metadata={"type": """list[str]"""})


@custom_define()
class ContentRepresentation(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        name: _No description given by bungie._
        path: _No description given by bungie._
        validation_string: _No description given by bungie._
    """

    name: str = custom_field()
    path: str = custom_field()
    validation_string: str = custom_field()


@custom_define()
class CommentSummary(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        comment_count: _No description given by bungie._
        topic_id: _No description given by bungie._
    """

    comment_count: int = custom_field()
    topic_id: int = custom_field(metadata={"int64": True})


@custom_define()
class NewsArticleRssResponse(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        category_filter: _No description given by bungie._
        current_pagination_token: _No description given by bungie._
        news_articles: _No description given by bungie._
        next_pagination_token: _No description given by bungie._
        pager_action: _No description given by bungie._
        result_count_this_page: _No description given by bungie._
    """

    category_filter: str = custom_field()
    current_pagination_token: int = custom_field()
    news_articles: list["NewsArticleRssItem"] = custom_field(metadata={"type": """list[NewsArticleRssItem]"""})
    next_pagination_token: int = custom_field()
    pager_action: str = custom_field()
    result_count_this_page: int = custom_field()


@custom_define()
class NewsArticleRssItem(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        description: _No description given by bungie._
        html_content: _No description given by bungie._
        image_path: _No description given by bungie._
        link: _No description given by bungie._
        optional_mobile_image_path: _No description given by bungie._
        pub_date: _No description given by bungie._
        title: _No description given by bungie._
        unique_identifier: _No description given by bungie._
    """

    description: str = custom_field()
    html_content: str = custom_field()
    image_path: str = custom_field()
    link: str = custom_field()
    optional_mobile_image_path: str = custom_field()
    pub_date: datetime = custom_field()
    title: str = custom_field()
    unique_identifier: str = custom_field()



import asyncio
import datetime
from typing import TYPE_CHECKING, Optional, Type

from sqlalchemy import JSON, Column, Table, Text, select, text
from sqlalchemy.exc import MultipleResultsFound
from sqlalchemy.ext.asyncio import AsyncConnection
from sqlalchemy.sql import Delete, Insert
from sqlalchemy.sql.ddl import CreateTable, DropTable

from bungio.http.route import Route
from bungio.models.base import ClientMixin, custom_define, custom_field
from bungio.utils import get_now_with_tz, split_list

if TYPE_CHECKING:
    from bungio.models.base import ManifestModel

__all__ = ("Manifest",)


@custom_define()
class Manifest(ClientMixin):
    """
    The connector to all manifest data
    """

    prefix: str = custom_field(init=False, default="destiny_manifest_")

    __synchronised: bool = custom_field(init=False, default=False)
    __saved_manifests: dict[str, Table] = custom_field(init=False, factory=dict)
    __manifest_urls: dict[str, str] = custom_field(init=False, factory=dict)
    __locks: dict[str, asyncio.Lock] = custom_field(init=False, factory=dict)
    __manifest_lock: asyncio.Lock = custom_field(init=False, default=asyncio.Lock())
    __synchronise_lock: asyncio.Lock = custom_field(init=False, default=asyncio.Lock())
    __manifest_last_update: datetime.datetime = custom_field(init=False, default=None)
    __version_table: Table = custom_field(init=False)

    def __attrs_post_init__(self):
        # noinspection PyProtectedMember
        self.__version_table = Table(
            f"{self.prefix}version",
            self._client._metadata,
            Column("version", Text, nullable=False, primary_key=True),
        )

    async def _synchronise_with_db(self):
        """
        Synchronise this client with the tables existing in the db
        """

        async with self.__synchronise_lock:
            if self.__synchronised:
                return

            db: AsyncConnection
            async with self._client.manifest_storage.begin() as db:
                # noinspection PyProtectedMember
                await db.run_sync(self._client._metadata.reflect)

                # create version table if that does not exist
                await db.execute(CreateTable(self.__version_table, if_not_exists=True))

                # noinspection PyProtectedMember
                # get all existing db tables
                for table_name, table in self._client._metadata.tables.items():
                    if table_name.startswith(self.prefix) and table_name != f"{self.prefix}version":
                        self.__saved_manifests[table_name.removeprefix(self.prefix)] = table

            await self._check_for_updates()
            self.__synchronised = True

    async def fetch(self, manifest_class: Type["ManifestModel"], value: str) -> Optional["ManifestModel"]:
        """
        Find a manifest model in the database by the reference, usually an id

        Args:
            manifest_class: The class the value should be converted to
            value: The value that references a manifest entry

        Returns:
            The manifest information, if found
        """

        db: AsyncConnection
        name = manifest_class.__name__

        if not self.__synchronised:
            await self._synchronise_with_db()

        await self.download(manifest_class=manifest_class)

        # get the data
        async with self._client.manifest_storage.begin() as db:
            query = select(self.__saved_manifests[name].columns.data).filter(
                self.__saved_manifests[name].columns.reference_id == str(value)
            )
            result = await db.execute(query)
            try:
                result = result.scalars().one_or_none()
            except MultipleResultsFound:
                return None
            if not result:
                return None

            return await manifest_class.from_dict(data=result, client=self._client, recursive=True)

    async def fetch_all(
        self, manifest_class: Type["ManifestModel"], filter: Optional[str] = None
    ) -> list["ManifestModel"]:
        """
        Return all models for a specific manifest entry

        Args:
            manifest_class: The classes to find in the manifest
            filter: A sql where clause with which you can filter the results. Example: `filter="CAST(data ->> 'itemType' AS INTEGER) = 3"` to get only weapons from `DestinyInventoryItemDefinition`

        Returns:
            The manifest information, if found
        """

        db: AsyncConnection
        name = manifest_class.__name__

        if not self.__synchronised:
            await self._synchronise_with_db()

        await self.download(manifest_class=manifest_class)

        # get the data
        async with self._client.manifest_storage.begin() as db:
            query = select(self.__saved_manifests[name].columns.data)
            if filter:
                query = query.filter(text(filter))

            results = await db.execute(query)
            results = results.scalars().all()

            return [
                await manifest_class.from_dict(data=result, client=self._client, recursive=True) for result in results
            ]

    async def download(self, manifest_class: Type["ManifestModel"] | str):
        """
        Download and save the manifest with the given name

        Args:
            manifest_class: The class / name of the manifest
        """

        if not isinstance(manifest_class, str):
            manifest_class = manifest_class.__name__

        # lock this
        if manifest_class not in self.__locks:
            self.__locks[manifest_class] = asyncio.Lock()

        async with self.__locks[manifest_class]:
            await self._check_for_updates()

            # create the table if that does not exist yet
            if manifest_class not in self.__saved_manifests:
                async with self._client.manifest_storage.begin() as db:
                    # noinspection PyProtectedMember
                    self.__saved_manifests[manifest_class] = Table(
                        f"{self.prefix}{manifest_class}",
                        self._client._metadata,
                        Column("reference_id", Text, nullable=False, primary_key=True),
                        Column("data", JSON, nullable=False),
                        keep_existing=True,
                    )
                    await db.execute(CreateTable(self.__saved_manifests[manifest_class], if_not_exists=True))

                    # fill the table
                    raw_data = await self._client.http.request(
                        Route(path=self.__manifest_urls[manifest_class], method="GET")
                    )
                    to_insert = [{"reference_id": str(key), "data": data} for key, data in raw_data.items()]

                    for part in split_list(to_split=to_insert, chunk_size=500):
                        await db.execute(Insert(self.__saved_manifests[manifest_class], values=part))

    async def _check_for_updates(self):
        """
        Checks if there is an updated version of the manifest available (hourly)

        Tip: Staying up to date
            This dispatches the `Client.on_manifest_update()` event
        """

        now = get_now_with_tz()
        async with self.__manifest_lock:
            if (
                self.__manifest_last_update is None
                or self.__manifest_urls == {}
                or self.__manifest_last_update < (now - datetime.timedelta(hours=1))
            ):
                manifest = await self._client.api.get_destiny_manifest()

                # set the urls
                self.__manifest_urls = {}
                for name, url in manifest.json_world_component_content_paths[self._client.language.value].items():
                    self.__manifest_urls[name] = url

                async with self._client.manifest_storage.begin() as db:
                    res = await db.execute(self.__version_table.select())
                    version = res.scalars().first()

                    if not version or version != manifest.version:
                        # delete tables
                        for table in self.__saved_manifests.values():
                            await db.execute(DropTable(table))
                        self.__saved_manifests = {}

                        # update version
                        await db.execute(Delete(self.__version_table))
                        await db.execute(Insert(self.__version_table, values={"version": manifest.version}))

                        # dispatch the update event
                        task = asyncio.create_task(self._client.on_manifest_update())
                        self._client._tasks.add(task)  # noqa
                        task.add_done_callback(self._client._tasks.discard)  # noqa

            self.__manifest_last_update = now



import asyncio
from asyncio import Semaphore
from typing import TYPE_CHECKING, Optional
from urllib.parse import urlencode

from aiohttp import (
    ClientConnectorCertificateError,
    ClientOSError,
    ClientResponse,
    ClientSession,
    DummyCookieJar,
    ServerDisconnectedError,
)

from bungio.error import (
    AuthenticationTooSlow,
    BadRequest,
    BungieDead,
    BungieException,
    InvalidAuthentication,
    NotFound,
    TimeoutException,
    _InvalidAuthentication,
    _RouteError,
)
from bungio.http.auth import AuthHttpRequests
from bungio.http.ratelimiting import RateLimiter
from bungio.http.route import Route
from bungio.http.routes import AllRouteHttpRequests
from bungio.models.auth import AuthData
from bungio.models.base import MISSING, ClientMixin, custom_define, custom_field
from bungio.singleton import Singleton

if TYPE_CHECKING:
    from aiohttp_client_cache import CachedSession

__all__ = ("HttpClient",)


@custom_define(init=False)
class HttpClient(AllRouteHttpRequests, AuthHttpRequests, ClientMixin, Singleton):
    """
    The singleton http client doing all communication with bungie

    Attributes:
        bungie_headers: Headers to use for bungie requests
        bungie_auth_headers: Headers to use for bungie auth requests
        max_attempts: Max retries before failure
    """

    _bungie_headers: dict[str, str] = custom_field()
    _bungie_auth_headers: dict[str, str] = custom_field()
    _max_attempts: int = custom_field(default=5)

    ratelimiter: RateLimiter = custom_field(init=False, default=RateLimiter())
    semaphore: Semaphore = custom_field(init=False, default=Semaphore(100))
    _session: "CachedSession | ClientSession" = custom_field(init=False, default=None)

    _initialised: bool = custom_field(init=False, default=False)

    def __init__(self, *args, **kwargs):
        if not getattr(self, "_initialised", False):
            self.__attrs_init__(*args, **kwargs)
        self._initialised = True

    def __del__(self):
        # clean up the session
        if self._session is not None:
            # noinspection PyProtectedMember
            self._session._connector.close()

    @property
    def session(self) -> "CachedSession | ClientSession":
        """Get the aiohttp session"""

        if not self._session:
            self.__setup_session()
        return self._session

    async def request(self, route: Route) -> dict:
        """
        Handle all web requests.

        Args:
            route: The route / method / params the request should have

        Raises:
            NotFound: 404 request
            BadRequest: 400 request
            InvalidAuthentication: If authentication is invalid
            TimeoutException: If no connection could be made
            BungieDead: Servers are down
            AuthenticationTooSlow: The authentication key has expired
            BungieException: Relaying the bungie error

        Returns:
            The json response
        """

        # make sure all is set up
        if not hasattr(self, "_client"):
            raise ValueError("You have to instantiate the `bungio.Client` before using this")

        headers = self._bungie_headers if not route.auth else await self._get_auth_headers(auth=route.auth)

        async with self.semaphore:
            return await self._request(
                route=route.path,
                method=route.method,
                headers=headers,
                params=route.params,
                data=route.data,
                auth=route.auth,
                use_ratelimiter=True,
            )

    async def _request(
        self,
        route: str,
        method: str,
        headers: dict,
        params: Optional[dict] = None,
        data: Optional[dict | list] = None,
        form_data: Optional[dict] = None,
        auth: Optional[AuthData] = None,
        use_ratelimiter: bool = True,
    ) -> dict:
        """
        Internal request function. Use HttpClient.request() instead!

        Args:
            route: The route
            method: The method
            headers: The headers
            params: The query params
            data: The body data
            form_data: The form data
            use_ratelimiter: If the ratelimiter should be used

        Raises:
            NotFound: 404 request
            BadRequest: 400 request
            InvalidAuthentication: If authentication is invalid
            TimeoutException: If no connection could be made
            BungieDead: Servers are down
            AuthenticationTooSlow: The authentication key has expired
            BungieException: Relaying the bungie error
            _RouteError: Getting redirected to a different url

        Returns:
            The json response.
        """

        route_with_params = f"{route}?{urlencode(params or {})}"

        for attempt in range(self._max_attempts):
            # wait for a token
            if use_ratelimiter:
                await self.ratelimiter.wait_for_token()

            try:
                async with self.session.request(
                    method=method,
                    url=route,
                    headers=headers,
                    params=params,
                    json=data,
                    data=form_data,
                ) as response:
                    self._client.logger.debug(f"[{response.status}] - {route_with_params}")

                    # cached responses do not have the content type field
                    content_type = getattr(response, "content_type", None) or response.headers.get(
                        "Content-Type", "NOT SET"
                    )

                    # get the json
                    content = (
                        await response.json(loads=self._client.json_loads) if "application/json" in content_type else {}
                    )

                    # make sure the response is not None
                    if content.get("Response", MISSING) is None:
                        content["Response"] = {}

                    if not await self._handle_response(
                        route_with_params=route_with_params, response=response, content=content
                    ):
                        continue

                    return content

            except _RouteError as e:
                route = e.route

            except _InvalidAuthentication:
                # noinspection PyProtectedMember
                self._client._invalidate_token(auth=auth)
                raise InvalidAuthentication(auth=auth)

            except (
                asyncio.exceptions.TimeoutError,
                ConnectionResetError,
                ServerDisconnectedError,
                ClientConnectorCertificateError,
                ClientOSError,
            ):
                self._client.logger.debug(
                    f"Retrying... - Timeout error for `{route}?{urlencode({} if params is None else params)}`"
                )
                await asyncio.sleep(2 + attempt * 3)
                continue

        self._client.logger.exception(
            f"{route_with_params}- Aborting. Failed {self._max_attempts} times. Infos: {method=} | {auth=} | {data=} | {form_data=}"
        )
        raise TimeoutException

    async def _handle_response(self, route_with_params: str, response: ClientResponse, content: dict) -> bool:
        """
        Handle the response returned by bungie

        Args:
            route_with_params: The full route for logging’s sake
            response: The response
            content: The json content returned or an empty dict

        Raises:
            NotFound: 404 request
            BadRequest: 400 request
            BungieDead: Servers are down
            AuthenticationTooSlow: The authentication key has expired
            BungieException: Relaying the bungie error
            _InvalidAuthentication: If authentication is invalid
            _RouteError: Getting redirected to a different url

        Returns:
            If the response was OK
        """

        # get the bungie errors from the json
        if (error := content.get("error", None)) is None:
            error = content.get("ErrorStatus", "MISSING")
            error_message = content.get("Message", "MISSING")
        else:
            # auth routes have different error code locations which are undocumented
            error_message = content.get("error_description", "MISSING")
            if error == "server_error":
                error = content.get("error_description", error)

        error_code = content.get("ErrorCode", -1)
        error_data = content.get("MessageData", {})

        match (response.status, error):
            case (_, "SystemDisabled"):
                raise BungieDead

            case (_, "invalid_grant"):
                # unauthorized
                self._client.logger.warning(
                    f"`{response.status} - {error} | {error_code}`: Unauthorized (too slow, user fault) request for `{route_with_params}`\n{content=}"
                )
                raise AuthenticationTooSlow

            case (200, _):
                # make sure we got a json
                if not content:
                    self._client.logger.warning(f"Wrong content type returned text: '{await response.text()}'")
                    await asyncio.sleep(3)
                    return False
                return True

            case (301, _):
                # this is issued with a 301 when bungie is of the opinion that resource this moved
                # just retrying fixes it with the new url fixes it
                new_route = response.headers.get("Location")
                self._client.logger.debug(
                    f"Wrong location, retrying with the correct one: `{response.url}` -> `{new_route}`"
                )
                raise _RouteError(route=new_route)

            case (502, _):
                # bad gateway
                self._client.logger.warning(
                    f"Retrying... - `{response.status}` Bad gateway error for `{route_with_params}`"
                )
                await asyncio.sleep(5)

            case (524, _):
                # cloudflare error, retry
                self._client.logger.warning(
                    f"Retrying... - `{response.status}` Cloudflare error for `{route_with_params}`"
                )
                await asyncio.sleep(2)

            case (429, _) | (_, "PerEndpointRequestThrottleExceeded" | "DestinyDirectBabelClientTimeout"):
                # we are getting throttled (should never be called in theory)
                self._client.logger.warning(
                    f"`{response.status} - {error} | {error_code}`: Retrying... - Getting throttled for `{route_with_params}`\n{content=}"
                )

                throttle_seconds = content["ThrottleSeconds"]

                # reset the ratelimit giver
                self.ratelimiter.tokens = 0
                await asyncio.sleep(max(throttle_seconds, 2))

            case (_, "DestinyDirectBabelClientTimeout"):
                # timeout
                self._client.logger.warning(
                    f"`{response.status} - {error} | {error_code}`: Retrying... - Getting timeouts for `{route_with_params}`\n{content=}"
                )
                await asyncio.sleep(60)

            case (_, "DestinyServiceFailure" | "DestinyInternalError" | "UnhandledException"):
                # timeout
                self._client.logger.warning(
                    f"`{response.status} - {error} | {error_code}`: Retrying... - Bungie is having problems `{route_with_params}`\n{content=}"
                )
                await asyncio.sleep(60)

            case (
                _,
                "AuthorizationRecordRevoked"
                | "AuthorizationRecordExpired",
            ):
                # users tokens are no longer valid
                raise _InvalidAuthentication()

            case (404, _):
                # not found
                self._client.logger.warning(
                    f"`{response.status} - {error} | {error_code}`: Not found for `{route_with_params}`\n{content=}"
                )

                if error != "MISSING":
                    raise BungieException(error=error, message=error_message, code=error_code, data=error_data)
                else:
                    raise NotFound

            case ((400 | 401), _):
                # generic bad request, such as wrong format
                self._client.logger.warning(
                    f"`{response.status} - {error} | {error_code}`: Generic bad request for `{route_with_params}`\n{content=}"
                )

                if error != "MISSING":
                    raise BungieException(error=error, message=error_message, code=error_code, data=error_data)
                else:
                    raise BadRequest

            case (503, _):
                self._client.logger.warning(
                    f"`{response.status} - {error} | {error_code}`: Retrying... - Server is overloaded for `{route_with_params}`\n{content=}"
                )
                await asyncio.sleep(10)

            case (status, _):
                # retry if we didn't get an error message
                retry = error == "Success"

                # catch the rest
                self._client.logger.warning(
                    f"`{status} - {error} | {error_code}` - Request failed for `{route_with_params}` - `{error_message}`"
                )

                if retry:
                    await asyncio.sleep(2)
                else:
                    raise BungieException(error=error, message=error_message, code=error_code, data=error_data)

        return False

    async def _get_auth_headers(self, auth: AuthData) -> dict:
        """
        Update the auth headers to include a working token. Raise an error if that doesn't exist

        Args:
            auth: The potentially old authentication info.

        Raises:
            InvalidAuthentication: If authentication is invalid

        Returns:
            User specific headers
        """

        headers = self._bungie_headers.copy()

        # get a working token or abort
        await self._client.get_working_auth(auth=auth)

        headers.update(
            {
                "Authorization": f"Bearer {auth.token}",
            }
        )

        return headers

    def __setup_session(self) -> None:
        """
        Sets up the session to use for http requests
        """

        try:
            from aiohttp_client_cache import CachedSession

            self._session = CachedSession(
                json_serialize=self._client.json_dumps, cookie_jar=DummyCookieJar(), cache=self._client.cache
            )
        except ModuleNotFoundError:
            self._session = ClientSession(json_serialize=self._client.json_dumps, cookie_jar=DummyCookieJar())



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import Optional, Union, TYPE_CHECKING

from bungio.utils import enum_converter
from bungio.models.base import BaseEnum, ManifestModel, custom_define, custom_field


if TYPE_CHECKING:
    from bungio.models import DestinyMedalTierDefinition


class DestinyActivityModeType(BaseEnum):
    """
    For historical reasons, this list will have both D1 and D2-relevant Activity Modes in it. Please don't take this to mean that some D1-only feature is coming back!
    """

    NONE = 0
    """_No description given by bungie._ """
    STORY = 2
    """_No description given by bungie._ """
    STRIKE = 3
    """_No description given by bungie._ """
    RAID = 4
    """_No description given by bungie._ """
    ALL_PV_P = 5
    """_No description given by bungie._ """
    PATROL = 6
    """_No description given by bungie._ """
    ALL_PV_E = 7
    """_No description given by bungie._ """
    RESERVED9 = 9
    """_No description given by bungie._ """
    CONTROL = 10
    """_No description given by bungie._ """
    RESERVED11 = 11
    """_No description given by bungie._ """
    CLASH = 12
    """Clash -> Destiny's name for Team Deathmatch. 4v4 combat, the team with the highest kills at the end of time wins. """
    RESERVED13 = 13
    """_No description given by bungie._ """
    CRIMSON_DOUBLES = 15
    """_No description given by bungie._ """
    NIGHTFALL = 16
    """_No description given by bungie._ """
    HEROIC_NIGHTFALL = 17
    """_No description given by bungie._ """
    ALL_STRIKES = 18
    """_No description given by bungie._ """
    IRON_BANNER = 19
    """_No description given by bungie._ """
    RESERVED20 = 20
    """_No description given by bungie._ """
    RESERVED21 = 21
    """_No description given by bungie._ """
    RESERVED22 = 22
    """_No description given by bungie._ """
    RESERVED24 = 24
    """_No description given by bungie._ """
    ALL_MAYHEM = 25
    """_No description given by bungie._ """
    RESERVED26 = 26
    """_No description given by bungie._ """
    RESERVED27 = 27
    """_No description given by bungie._ """
    RESERVED28 = 28
    """_No description given by bungie._ """
    RESERVED29 = 29
    """_No description given by bungie._ """
    RESERVED30 = 30
    """_No description given by bungie._ """
    SUPREMACY = 31
    """_No description given by bungie._ """
    PRIVATE_MATCHES_ALL = 32
    """_No description given by bungie._ """
    SURVIVAL = 37
    """_No description given by bungie._ """
    COUNTDOWN = 38
    """_No description given by bungie._ """
    TRIALS_OF_THE_NINE = 39
    """_No description given by bungie._ """
    SOCIAL = 40
    """_No description given by bungie._ """
    TRIALS_COUNTDOWN = 41
    """_No description given by bungie._ """
    TRIALS_SURVIVAL = 42
    """_No description given by bungie._ """
    IRON_BANNER_CONTROL = 43
    """_No description given by bungie._ """
    IRON_BANNER_CLASH = 44
    """_No description given by bungie._ """
    IRON_BANNER_SUPREMACY = 45
    """_No description given by bungie._ """
    SCORED_NIGHTFALL = 46
    """_No description given by bungie._ """
    SCORED_HEROIC_NIGHTFALL = 47
    """_No description given by bungie._ """
    RUMBLE = 48
    """_No description given by bungie._ """
    ALL_DOUBLES = 49
    """_No description given by bungie._ """
    DOUBLES = 50
    """_No description given by bungie._ """
    PRIVATE_MATCHES_CLASH = 51
    """_No description given by bungie._ """
    PRIVATE_MATCHES_CONTROL = 52
    """_No description given by bungie._ """
    PRIVATE_MATCHES_SUPREMACY = 53
    """_No description given by bungie._ """
    PRIVATE_MATCHES_COUNTDOWN = 54
    """_No description given by bungie._ """
    PRIVATE_MATCHES_SURVIVAL = 55
    """_No description given by bungie._ """
    PRIVATE_MATCHES_MAYHEM = 56
    """_No description given by bungie._ """
    PRIVATE_MATCHES_RUMBLE = 57
    """_No description given by bungie._ """
    HEROIC_ADVENTURE = 58
    """_No description given by bungie._ """
    SHOWDOWN = 59
    """_No description given by bungie._ """
    LOCKDOWN = 60
    """_No description given by bungie._ """
    SCORCHED = 61
    """_No description given by bungie._ """
    SCORCHED_TEAM = 62
    """_No description given by bungie._ """
    GAMBIT = 63
    """_No description given by bungie._ """
    ALL_PV_E_COMPETITIVE = 64
    """_No description given by bungie._ """
    BREAKTHROUGH = 65
    """_No description given by bungie._ """
    BLACK_ARMORY_RUN = 66
    """_No description given by bungie._ """
    SALVAGE = 67
    """_No description given by bungie._ """
    IRON_BANNER_SALVAGE = 68
    """_No description given by bungie._ """
    PV_P_COMPETITIVE = 69
    """_No description given by bungie._ """
    PV_P_QUICKPLAY = 70
    """_No description given by bungie._ """
    CLASH_QUICKPLAY = 71
    """_No description given by bungie._ """
    CLASH_COMPETITIVE = 72
    """_No description given by bungie._ """
    CONTROL_QUICKPLAY = 73
    """_No description given by bungie._ """
    CONTROL_COMPETITIVE = 74
    """_No description given by bungie._ """
    GAMBIT_PRIME = 75
    """_No description given by bungie._ """
    RECKONING = 76
    """_No description given by bungie._ """
    MENAGERIE = 77
    """_No description given by bungie._ """
    VEX_OFFENSIVE = 78
    """_No description given by bungie._ """
    NIGHTMARE_HUNT = 79
    """_No description given by bungie._ """
    ELIMINATION = 80
    """_No description given by bungie._ """
    MOMENTUM = 81
    """_No description given by bungie._ """
    DUNGEON = 82
    """_No description given by bungie._ """
    SUNDIAL = 83
    """_No description given by bungie._ """
    TRIALS_OF_OSIRIS = 84
    """_No description given by bungie._ """
    DARES = 85
    """_No description given by bungie._ """
    OFFENSIVE = 86
    """_No description given by bungie._ """
    LOST_SECTOR = 87
    """_No description given by bungie._ """
    RIFT = 88
    """_No description given by bungie._ """
    ZONE_CONTROL = 89
    """_No description given by bungie._ """
    IRON_BANNER_RIFT = 90
    """_No description given by bungie._ """
    IRON_BANNER_ZONE_CONTROL = 91
    """_No description given by bungie._ """
    RELIC = 92
    """_No description given by bungie._ """


@custom_define()
class DestinyHistoricalStatsDefinition(ManifestModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        category: Category for the stat.
        group: Statistic group
        icon_image: Optional URI to an icon for the statistic
        medal_tier_hash: The tier associated with this medal - be it implicitly or explicitly.
        merge_method: Optional icon for the statistic
        modes: Game modes where this statistic can be reported.
        period_types: Time periods the statistic covers
        stat_description: Description of a stat if applicable.
        stat_id: Unique programmer friendly ID for this stat
        stat_name: Display name
        stat_name_abbr: Display name abbreviated
        unit_label: Localized Unit Name for the stat.
        unit_type: Unit, if any, for the statistic
        weight: Weight assigned to this stat indicating its relative impressiveness.
        manifest_medal_tier_hash: Manifest information for `medal_tier_hash`
    """

    category: Union["DestinyStatsCategoryType", int] = custom_field(
        converter=enum_converter("DestinyStatsCategoryType")
    )
    group: Union["DestinyStatsGroupType", int] = custom_field(converter=enum_converter("DestinyStatsGroupType"))
    icon_image: str = custom_field()
    medal_tier_hash: int = custom_field()
    merge_method: int = custom_field()
    modes: list[Union["DestinyActivityModeType", int]] = custom_field(
        converter=enum_converter("DestinyActivityModeType")
    )
    period_types: list[Union["PeriodType", int]] = custom_field(converter=enum_converter("PeriodType"))
    stat_description: str = custom_field()
    stat_id: str = custom_field()
    stat_name: str = custom_field()
    stat_name_abbr: str = custom_field()
    unit_label: str = custom_field()
    unit_type: Union["UnitType", int] = custom_field(converter=enum_converter("UnitType"))
    weight: int = custom_field()
    manifest_medal_tier_hash: Optional["DestinyMedalTierDefinition"] = custom_field(default=None)


class DestinyStatsGroupType(BaseEnum):
    """
    If the enum value is > 100, it is a "special" group that cannot be queried for directly (special cases apply to when they are returned, and are not relevant in general cases)
    """

    NONE = 0
    """_No description given by bungie._ """
    GENERAL = 1
    """_No description given by bungie._ """
    WEAPONS = 2
    """_No description given by bungie._ """
    MEDALS = 3
    """_No description given by bungie._ """
    RESERVED_GROUPS = 100
    """This is purely to serve as the dividing line between filterable and un-filterable groups. Below this number is a group you can pass as a filter. Above it are groups used in very specific circumstances and not relevant for filtering. """
    LEADERBOARD = 101
    """Only applicable while generating leaderboards. """
    ACTIVITY = 102
    """These will *only* be consumed by GetAggregateStatsByActivity """
    UNIQUE_WEAPON = 103
    """These are only consumed and returned by GetUniqueWeaponHistory """
    INTERNAL = 104
    """_No description given by bungie._ """


class DestinyStatsCategoryType(BaseEnum):
    """
    _No description given by bungie._
    """

    NONE = 0
    """_No description given by bungie._ """
    KILLS = 1
    """_No description given by bungie._ """
    ASSISTS = 2
    """_No description given by bungie._ """
    DEATHS = 3
    """_No description given by bungie._ """
    CRITICALS = 4
    """_No description given by bungie._ """
    K_DA = 5
    """_No description given by bungie._ """
    K_D = 6
    """_No description given by bungie._ """
    SCORE = 7
    """_No description given by bungie._ """
    ENTERED = 8
    """_No description given by bungie._ """
    TIME_PLAYED = 9
    """_No description given by bungie._ """
    MEDAL_WINS = 10
    """_No description given by bungie._ """
    MEDAL_GAME = 11
    """_No description given by bungie._ """
    MEDAL_SPECIAL_KILLS = 12
    """_No description given by bungie._ """
    MEDAL_SPREES = 13
    """_No description given by bungie._ """
    MEDAL_MULTI_KILLS = 14
    """_No description given by bungie._ """
    MEDAL_ABILITIES = 15
    """_No description given by bungie._ """


class UnitType(BaseEnum):
    """
    _No description given by bungie._
    """

    NONE = 0
    """_No description given by bungie._ """
    COUNT = 1
    """Indicates the statistic is a simple count of something. """
    PER_GAME = 2
    """Indicates the statistic is a per game average. """
    SECONDS = 3
    """Indicates the number of seconds """
    POINTS = 4
    """Indicates the number of points earned """
    TEAM = 5
    """Values represents a team ID """
    DISTANCE = 6
    """Values represents a distance (units to-be-determined) """
    PERCENT = 7
    """Ratio represented as a whole value from 0 to 100. """
    RATIO = 8
    """Ratio of something, shown with decimal places """
    BOOLEAN = 9
    """True or false """
    WEAPON_TYPE = 10
    """The stat is actually a weapon type. """
    STANDING = 11
    """Indicates victory, defeat, or something in between. """
    MILLISECONDS = 12
    """Number of milliseconds some event spanned. For example, race time, or lap time. """
    COMPLETION_REASON = 13
    """The value is a enumeration of the Completion Reason type. """


class DestinyStatsMergeMethod(BaseEnum):
    """
    _No description given by bungie._
    """

    ADD = 0
    """When collapsing multiple instances of the stat together, add the values. """
    MIN = 1
    """When collapsing multiple instances of the stat together, take the lower value. """
    MAX = 2
    """When collapsing multiple instances of the stat together, take the higher value. """


class PeriodType(BaseEnum):
    """
    _No description given by bungie._
    """

    NONE = 0
    """_No description given by bungie._ """
    DAILY = 1
    """_No description given by bungie._ """
    ALL_TIME = 2
    """_No description given by bungie._ """
    ACTIVITY = 3
    """_No description given by bungie._ """



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import Optional, Union

from bungio.models.base import ClientMixin, custom_define
from bungio.models.auth import AuthData

from bungio.models import SearchResultOfFireteamResponse
from bungio.models import FireteamPlatform
from bungio.models import FireteamSlotSearch
from bungio.models import FireteamResponse
from bungio.models import FireteamDateRange
from bungio.models import SearchResultOfFireteamSummary
from bungio.models import FireteamPublicSearchOption


@custom_define()
class FireteamRouteInterface(ClientMixin):
    async def get_active_private_clan_fireteam_count(self, group_id: int, auth: AuthData) -> int:
        """
        Gets a count of all active non-public fireteams for the specified clan. Maximum value returned is 25.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadGroups

        Args:
            group_id: The group id of the clan.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_active_private_clan_fireteam_count(group_id=group_id, auth=auth)
        return response["Response"]

    async def get_available_clan_fireteams(
        self,
        activity_type: int,
        date_range: Union[FireteamDateRange, int],
        group_id: int,
        page: int,
        platform: Union[FireteamPlatform, int],
        public_only: Union[FireteamPublicSearchOption, int],
        slot_filter: Union[FireteamSlotSearch, int],
        auth: AuthData,
        exclude_immediate: Optional[bool] = None,
        lang_filter: Optional[str] = None,
    ) -> SearchResultOfFireteamSummary:
        """
        Gets a listing of all of this clan's fireteams that are have available slots. Caller is not checked for join criteria so caching is maximized.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadGroups

        Args:
            activity_type: The activity type to filter by.
            date_range: The date range to grab available fireteams.
            group_id: The group id of the clan.
            page: Zero based page
            platform: The platform filter.
            public_only: Determines public/private filtering.
            slot_filter: Filters based on available slots
            auth: Authentication information.
            exclude_immediate: If you wish the result to exclude immediate fireteams, set this to true. Immediate-only can be forced using the dateRange enum.
            lang_filter: An optional language filter.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_available_clan_fireteams(
            activity_type=activity_type,
            date_range=getattr(date_range, "value", date_range),
            group_id=group_id,
            page=page,
            platform=getattr(platform, "value", platform),
            public_only=getattr(public_only, "value", public_only),
            slot_filter=getattr(slot_filter, "value", slot_filter),
            auth=auth,
            exclude_immediate=exclude_immediate if exclude_immediate is not None else None,
            lang_filter=lang_filter if lang_filter is not None else None,
        )
        return await SearchResultOfFireteamSummary.from_dict(
            data=response,
            client=self._client,
            activity_type=activity_type,
            date_range=date_range,
            group_id=group_id,
            page=page,
            platform=platform,
            public_only=public_only,
            slot_filter=slot_filter,
            auth=auth,
            exclude_immediate=exclude_immediate,
            lang_filter=lang_filter,
        )

    async def search_public_available_clan_fireteams(
        self,
        activity_type: int,
        date_range: Union[FireteamDateRange, int],
        page: int,
        platform: Union[FireteamPlatform, int],
        slot_filter: Union[FireteamSlotSearch, int],
        auth: AuthData,
        exclude_immediate: Optional[bool] = None,
        lang_filter: Optional[str] = None,
    ) -> SearchResultOfFireteamSummary:
        """
        Gets a listing of all public fireteams starting now with open slots. Caller is not checked for join criteria so caching is maximized.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadGroups

        Args:
            activity_type: The activity type to filter by.
            date_range: The date range to grab available fireteams.
            page: Zero based page
            platform: The platform filter.
            slot_filter: Filters based on available slots
            auth: Authentication information.
            exclude_immediate: If you wish the result to exclude immediate fireteams, set this to true. Immediate-only can be forced using the dateRange enum.
            lang_filter: An optional language filter.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.search_public_available_clan_fireteams(
            activity_type=activity_type,
            date_range=getattr(date_range, "value", date_range),
            page=page,
            platform=getattr(platform, "value", platform),
            slot_filter=getattr(slot_filter, "value", slot_filter),
            auth=auth,
            exclude_immediate=exclude_immediate if exclude_immediate is not None else None,
            lang_filter=lang_filter if lang_filter is not None else None,
        )
        return await SearchResultOfFireteamSummary.from_dict(
            data=response,
            client=self._client,
            activity_type=activity_type,
            date_range=date_range,
            page=page,
            platform=platform,
            slot_filter=slot_filter,
            auth=auth,
            exclude_immediate=exclude_immediate,
            lang_filter=lang_filter,
        )

    async def get_my_clan_fireteams(
        self,
        group_id: int,
        include_closed: bool,
        page: int,
        platform: Union[FireteamPlatform, int],
        auth: AuthData,
        group_filter: Optional[bool] = None,
        lang_filter: Optional[str] = None,
    ) -> SearchResultOfFireteamResponse:
        """
        Gets a listing of all fireteams that caller is an applicant, a member, or an alternate of.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadGroups

        Args:
            group_id: The group id of the clan. (This parameter is ignored unless the optional query parameter groupFilter is true).
            include_closed: If true, return fireteams that have been closed.
            page: Deprecated parameter, ignored.
            platform: The platform filter.
            auth: Authentication information.
            group_filter: If true, filter by clan. Otherwise, ignore the clan and show all of the user's fireteams.
            lang_filter: An optional language filter.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_my_clan_fireteams(
            group_id=group_id,
            include_closed=include_closed,
            page=page,
            platform=getattr(platform, "value", platform),
            auth=auth,
            group_filter=group_filter if group_filter is not None else None,
            lang_filter=lang_filter if lang_filter is not None else None,
        )
        return await SearchResultOfFireteamResponse.from_dict(
            data=response,
            client=self._client,
            group_id=group_id,
            include_closed=include_closed,
            page=page,
            platform=platform,
            auth=auth,
            group_filter=group_filter,
            lang_filter=lang_filter,
        )

    async def get_clan_fireteam(self, fireteam_id: int, group_id: int, auth: AuthData) -> FireteamResponse:
        """
        Gets a specific fireteam.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadGroups

        Args:
            fireteam_id: The unique id of the fireteam.
            group_id: The group id of the clan.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_clan_fireteam(fireteam_id=fireteam_id, group_id=group_id, auth=auth)
        return await FireteamResponse.from_dict(
            data=response, client=self._client, fireteam_id=fireteam_id, group_id=group_id, auth=auth
        )



from datetime import datetime
from typing import Callable, Coroutine, Optional

from bungio.http.route import Route
from bungio.models.auth import AuthData


class AppRouteHttpRequests:
    request: Callable[..., Coroutine]

    async def get_application_api_usage(
        self,
        application_id: int,
        auth: AuthData,
        end: Optional[datetime] = None,
        start: Optional[datetime] = None,
        *args,
        **kwargs,
    ) -> dict:
        """
        Get API usage by application for time frame specified. You can go as far back as 30 days ago, and can ask for up to a 48 hour window of time in a single request. You must be authenticated with at least the ReadUserData permission to access this endpoint.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadUserData

        Args:
            application_id: ID of the application to get usage statistics.
            auth: Authentication information.
            end: End time for query. Goes to now if not specified.
            start: Start time for query. Goes to 24 hours ago if not specified.

        Raises:
            NotFound: 404 request
            BadRequest: 400 request
            InvalidAuthentication: If authentication is invalid
            TimeoutException: If no connection could be made
            BungieDead: Servers are down
            AuthenticationTooSlow: The authentication key has expired
            BungieException: Relaying the bungie error

        Returns:
            The json response
        """

        return await self.request(
            Route(path=f"/App/ApiUsage/{application_id}/", method="GET", auth=auth, end=end, start=start)
        )

    async def get_bungie_applications(self, auth: Optional[AuthData] = None, *args, **kwargs) -> dict:
        """
        Get list of applications created by Bungie.

        Args:
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Raises:
            NotFound: 404 request
            BadRequest: 400 request
            InvalidAuthentication: If authentication is invalid
            TimeoutException: If no connection could be made
            BungieDead: Servers are down
            AuthenticationTooSlow: The authentication key has expired
            BungieException: Relaying the bungie error

        Returns:
            The json response
        """

        return await self.request(Route(path="/App/FirstParty/", method="GET", auth=auth))
